from privacyidea.lib.container import init_container, add_token_to_container, find_container_by_serial
from privacyidea.lib.event import set_event, delete_event
from privacyidea.lib.eventhandler.containerhandler import (ContainerEventHandler, ACTION_TYPE as C_ACTION_TYPE)
from privacyidea.lib.eventhandler.customuserattributeshandler import ACTION_TYPE, USER_TYPE
from privacyidea.lib.policy import SCOPE, set_policy, delete_policy, ACTION
from privacyidea.lib.token import init_token, remove_token
from privacyidea.lib.user import User
from .base import MyApiTestCase, FakeFlaskG
from . import smtpmock
import mock
from privacyidea.lib.config import set_privacyidea_config
from .test_lib_events import ContainerEventTestCase
from .test_lib_tokencontainer import MockSmartphone

# TODO: this should be imported from lib.event when available
HANDLERS = ["UserNotification", "Token", "Federation", "Script", "Counter",
            "RequestMangler", "ResponseMangler", "Logging", "CustomUserAttributes", "Container"]


class APIEventsTestCase(MyApiTestCase):

    def test_00_api_errors(self):
        # check for auth error
        with self.app.test_request_context('/event/',
                                           method='GET'):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 401, res)
            self.assertEqual(res.json['result']['error']['code'], 4033, res.json)

        # check for automatic redirect on missing trailing slash from flask
        with self.app.test_request_context('/event',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            # In test environment this can throw a 301
            self.assertTrue(res.status_code in [301, 308], res)
            self.assertIn('Location', res.headers, res)

        param = {
            "name": "Send an email",
            "event": "token_init",
            "action": "sendmail",
            "handlermodule": "UserNotification",
            "conditions": '{"blabla": "yes"}'
        }
        with self.app.test_request_context('/event',
                                           method='POST',
                                           data=param,
                                           headers={'Authorization': self.at}):
            # check for policy error with a restrictive admin policy
            set_policy(name="adm_disable_event", scope=SCOPE.ADMIN,
                       action="configwrite")
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 403, res)
            self.assertEqual(res.json['result']['error']['code'], 303, res.json)
            delete_policy('adm_disable_event')

        # check fo resourceNotFound error
        with self.app.test_request_context('/event/enable/1234',
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertEqual(res.status_code, 404, res)
            self.assertEqual(res.json['result']['error']['code'], 601, res.json)
        # And check if the /event request writes a valid (failed) audit entry
        auditentry = self.find_most_recent_audit_entry(action='POST /event/enable/<eventid>')
        self.assertEqual(auditentry['action'], 'POST /event/enable/<eventid>', auditentry)
        self.assertEqual(auditentry['success'], 0, auditentry)

    def test_01_crud_events(self):
        # list empty events
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertEqual(result.get("value"), [])
            self.assertIn('signature', res.json, res.json)

        # Check if the /event request writes a valid audit entry
        auditentry = self.find_most_recent_audit_entry(action='GET /event/')
        self.assertEqual(auditentry['success'], 1, auditentry)

        # create an event configuration
        param = {
            "name": "Send an email",
            "event": "token_init",
            "action": "sendmail",
            "handlermodule": "UserNotification",
            "conditions": '{"blabla": "yes"}'
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # check the event
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value")[0].get("action"), "sendmail")
            self.assertEqual(result.get("value")[0].get("conditions"),
                             {"blabla": "yes"})

        # update event config
        param = {
            "name": "A new name",
            "event": "token_init",
            "action": "sendmail",
            "handlermodule": "UserNotification",
            "conditions": '{"always": "yes"}',
            "id": 1
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # check the event
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value")[0].get("action"),
                             "sendmail")
            self.assertEqual(result.get("value")[0].get("conditions"),
                             {"always": "yes"})

        # get one single event
        with self.app.test_request_context('/event/1',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value")[0].get("action"),
                             "sendmail")
            self.assertEqual(result.get("value")[0].get("conditions"),
                             {"always": "yes"})

        # delete event
        with self.app.test_request_context('/event/1',
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), 1)

        # list empty events
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

    def test_02_test_options(self):
        # create an event configuration
        param = {
            "name": "Send an email via themis",
            "event": "token_init",
            "action": "sendmail",
            "handlermodule": "UserNotification",
            "conditions": '{"blabla": "yes"}',
            "option.emailconfig": "themis",
            "option.2": "value2"
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertGreaterEqual(result.get("value"), 1, result)
            ev1_id = result.get('value')

        # list event with options
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            event_list = result.get("value")
            self.assertEqual(len(event_list), 1)
            self.assertEqual(event_list[0].get("action"), "sendmail")
            self.assertEqual(event_list[0].get("event"), ["token_init"])
            self.assertEqual(event_list[0].get("options").get("2"), "value2")
            self.assertEqual(event_list[0].get("options").get("emailconfig"),
                             "themis")

        # delete event
        with self.app.test_request_context('/event/{0!s}'.format(ev1_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), ev1_id)

        # list empty events
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

    def test_03_available_events(self):
        with self.app.test_request_context('/event/available',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertTrue("token_init" in result.get("value"))
            self.assertTrue("token_assign" in result.get("value"))
            self.assertTrue("token_unassign" in result.get("value"))

    def test_04_handler_modules(self):
        with self.app.test_request_context('/event/handlermodules',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            for h in HANDLERS:
                self.assertIn(h, result.get("value"), result)

    def test_05_get_handler_actions(self):
        with self.app.test_request_context('/event/actions/UserNotification',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("sendmail" in result.get("value"))
            self.assertTrue("sendsms" in result.get("value"))
            detail = res.json.get("detail")

        with self.app.test_request_context('/event/actions/Token',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            set_random_pin = result.get("value").get("set random pin")
            # The valid OTP PIN length is returned as list
            self.assertTrue(type(set_random_pin.get("length").get("value")), "list")

    def test_05_get_handler_conditions(self):
        with self.app.test_request_context('/event/conditions/UserNotification',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("logged_in_user" in result.get("value"))
            self.assertTrue("result_value" in result.get("value"))
            detail = res.json.get("detail")

    def test_06_test_enable_disable(self):
        # create an event configuration
        param = {
            "name": "Send an email via themis",
            "event": "token_init",
            "action": "sendmail",
            "handlermodule": "UserNotification",
            "conditions": '{"blabla": "yes"}',
            "option.emailconfig": "themis",
            "option.2": "value2"
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertGreaterEqual(result.get("value"), 1, result)
            ev1_id = result.get('value')

        # list event with options
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            event_list = result.get("value")
            self.assertEqual(len(event_list), 1)
            self.assertEqual(event_list[0].get("active"), True)

        # disable event
        with self.app.test_request_context('/event/disable/{0!s}'.format(ev1_id),
                                           method='POST',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            event_list = result.get("value")
            self.assertEqual(len(event_list), 1)
            self.assertEqual(event_list[0].get("active"), False)

        # Enable event
        with self.app.test_request_context('/event/enable/{0!s}'.format(ev1_id),
                                           method='POST',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)

        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            event_list = result.get("value")
            self.assertEqual(len(event_list), 1)
            self.assertEqual(event_list[0].get("active"), True)

        # delete event
        with self.app.test_request_context('/event/{0!s}'.format(ev1_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), ev1_id, result)

        # list empty events
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

    def test_07_positions(self):
        # test the available Positions

        with self.app.test_request_context('/event/positions/Token',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue("post" in result.get("value"), result.get("value"))
            self.assertTrue("pre" in result.get("value"), result.get("value"))

        # create an event configuration
        param = {
            "name": "Delete the token that was created :-)",
            "event": "token_init",
            "action": "delete",
            "handlermodule": "Token",
            "conditions": '{"blabla": "yes"}',
            "position": "post"
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertGreaterEqual(result.get("value"), 1, result)
            ev1_id = result.get('value')

        # check the event
        with self.app.test_request_context('/event/{0!s}'.format(ev1_id),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value")[0].get("position"), "post")

        # Update event with the position=pre
        param["id"] = ev1_id
        param["position"] = "pre"
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), ev1_id, result)

        # check the event
        with self.app.test_request_context('/event/{0!s}'.format(ev1_id),
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value")[0].get("position"), "pre")

        # delete event
        with self.app.test_request_context('/event/{0!s}'.format(ev1_id),
                                           method='DELETE',
                                           headers={
                                               'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), ev1_id, result)

        # list empty events
        with self.app.test_request_context('/event/',
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value"), [])

    @smtpmock.activate
    def test_08_create_token_for_user(self):
        smtpmock.setdata(response={"pi_tester@privacyidea.org": (200, 'OK')})
        transactionid = "123456098712"
        # send the email with the old configuration
        set_privacyidea_config("email.mailserver", "localhost")
        set_privacyidea_config("email.username", "user")
        set_privacyidea_config("email.username", "password")
        set_privacyidea_config("email.tls", True)
        # We create a token for a user, who has currently no token!
        # create an event configuration
        param = {
            "name": "Create Email Token for untokened user",
            "event": "validate_check",
            "action": "enroll",
            "handlermodule": "Token",
            "position": "pre",
            "conditions": '{"user_token_number": "0"}',
            "option.user": "true",
            "option.tokentype": "email",
            "option.dynamic_email": "1",
            "option.additional_params": "{'pin':'1234'}"
        }
        with self.app.test_request_context('/event',
                                           data=param,
                                           method='POST',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertGreaterEqual(result.get("value"), 1, result)

        self.setUp_user_realm2()
        # usernotoken, self.realm2
        # check that the user has no tokens
        with self.app.test_request_context('/token/',
                                           data={"user": "usernotoken",
                                                 "realm": self.realm2},
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value").get("tokens"), [])

        # user tries to authenticate. with the pin 1234.
        # He gets an email token enrolled and gets the transaction code in the response
        with self.app.test_request_context('/validate/check',
                                           data={"user": "usernotoken",
                                                 "realm": self.realm2,
                                                 "pass": "1234"},
                                           method="POST",
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertFalse(result.get("value"))
            self.assertTrue(detail.get("serial").startswith("PIEM"))
            # This is a challenge response!
            self.assertTrue("transaction_id" in detail, detail)
            self.assertTrue("multi_challenge" in detail, detail)
            self.assertEqual(detail.get("message"), "Enter the OTP from the Email")

        # check user has a token
        with self.app.test_request_context('/token/',
                                           data={"user": "usernotoken",
                                                 "realm": self.realm2},
                                           method='GET',
                                           headers={'Authorization': self.at}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            detail = res.json.get("detail")
            self.assertEqual(result.get("value").get("count"), 1)
            self.assertEqual(result.get("value").get("tokens")[0].get("tokentype"), "email")

    def test_09_get_container_serial(self):
        # Get container serial from response
        payload = {"type": "Smartphone", "description": "test description!!"}
        request = self.app.test_request_context('/container/init',
                                                method='POST',
                                                data=payload,
                                                headers={'Authorization': self.at})
        with request:
            res = self.app.full_dispatch_request()

            c_handler = ContainerEventHandler()
            content = c_handler._get_response_content(res)
            container_serial = c_handler._get_container_serial(request.request, content)

            self.assertIsNotNone(container_serial)

        # Get container serial from request
        container_serial = init_container({"type": "Generic", "description": "test description!!"})["container_serial"]
        payload = {"container_serial": container_serial, "states": ["active"]}
        request = self.app.test_request_context(f'/container/{container_serial}/states',
                                                method='POST',
                                                data=payload,
                                                headers={'Authorization': self.at})
        with request:
            res = self.app.full_dispatch_request()

            c_handler = ContainerEventHandler()
            container_serial_req = c_handler._get_container_serial(request.request, res)

            self.assertEqual(container_serial_req, container_serial)

        # Get container serial from token
        token_serial = "SPASS0001"
        init_token({"type": "spass", "serial": token_serial})
        add_token_to_container(container_serial, token_serial)

        payload = {"serial": token_serial}
        request = self.app.test_request_context('/token/enable',
                                                method='POST',
                                                data=payload,
                                                headers={'Authorization': self.at})
        with request:
            res = self.app.full_dispatch_request()
            g = FakeFlaskG()

            c_handler = ContainerEventHandler()
            container_serial_token = c_handler._get_container_serial_from_token(request.request, res, g)

            self.assertEqual(container_serial_token, container_serial)


class CustomUserAttributeHandlerTestCase(MyApiTestCase):
    def setUp(self):
        super(CustomUserAttributeHandlerTestCase, self).setUp()
        self.setUp_user_realms()

    def test_01_user_attribute_with_handler_tokenowner(self):
        user = User('cornelius', self.realm1)
        tok = init_token({'type': 'spass', 'pin': 'test'})
        self.assertNotIn('foo', user.attributes, user.attributes)

        # First try to delete a non-existing attribute
        eid = set_event("user_atts", event=["validate_check"],
                        action=ACTION_TYPE.DELETE_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.TOKENOWNER,
                                 'attrkey': 'foo'})

        # what happens if the token has no user
        with self.app.test_request_context('/validate/check',
                                           data={"serial": tok.token.serial,
                                                 "pass": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"), result)
            self.assertNotIn('foo', user.attributes, user.attributes)

        # now we attach the token to a user
        tok.add_user(user)
        with self.app.test_request_context('/validate/check',
                                           data={"user": 'cornelius',
                                                 "pass": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"), result)
            self.assertNotIn('foo', user.attributes, user.attributes)

        # update event to add an attribute
        eid = set_event("user_atts", event=["validate_check"], id=eid,
                        action=ACTION_TYPE.SET_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.TOKENOWNER,
                                 'attrkey': 'foo',
                                 'attrvalue': 'bar'})
        with self.app.test_request_context('/validate/check',
                                           data={"user": "cornelius",
                                                 "pass": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"), result)
            self.assertEqual('bar', user.attributes.get('foo'), user.attributes)

        # update event to update an attribute
        eid = set_event("user_atts", event=["validate_check"], id=eid,
                        action=ACTION_TYPE.SET_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.TOKENOWNER,
                                 'attrkey': 'foo',
                                 'attrvalue': 'baz'})
        with self.app.test_request_context('/validate/check',
                                           data={"user": "cornelius",
                                                 "pass": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"), result)
            self.assertEqual('baz', user.attributes.get('foo'), user.attributes)

        # now delete the added attribute
        eid = set_event("user_atts", event=["validate_check"], id=eid,
                        action=ACTION_TYPE.DELETE_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.TOKENOWNER,
                                 'attrkey': 'foo'})
        with self.app.test_request_context('/validate/check',
                                           data={"user": "cornelius",
                                                 "pass": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            result = res.json.get("result")
            self.assertTrue(result.get("value"), result)
            self.assertNotIn('foo', user.attributes, user.attributes)

        delete_event(eid)
        remove_token(tok.token.serial)

    def test_02_user_attribute_with_handler_logged_in_user(self):
        user = User('cornelius', realm=self.realm1)
        self.assertNotIn('foo', user.attributes, user.attributes)

        # get the auth-token for the user
        with self.app.test_request_context('/auth',
                                           data={"username": 'cornelius',
                                                 "password": 'test'},
                                           method='POST'):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res)
            result = res.json.get("result")
            self.assertTrue(result.get("status"), result)
            user_token = result.get("value").get("token")

        # try to delete a non-existing attribute
        eid = set_event("user_atts", event=["token_list"],
                        action=ACTION_TYPE.DELETE_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.LOGGED_IN_USER,
                                 'attrkey': 'foo'})
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': user_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertNotIn('foo', user.attributes, user.attributes)

        # delete an existing attribute
        user.set_attribute('foo', 'bar')
        self.assertIn('foo', user.attributes, user.attributes)
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': user_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertNotIn('foo', user.attributes, user.attributes)

        # add an attribute
        eid = set_event("user_atts", event=["token_list"], id=eid,
                        action=ACTION_TYPE.SET_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.LOGGED_IN_USER,
                                 'attrkey': 'foo',
                                 'attrvalue': 'bar'})
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': user_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual('bar', user.attributes['foo'], user.attributes)

        # overwrite an attribute
        eid = set_event("user_atts", event=["token_list"], id=eid,
                        action=ACTION_TYPE.SET_CUSTOM_USER_ATTRIBUTES,
                        handlermodule="CustomUserAttributes", conditions={},
                        options={'user': USER_TYPE.LOGGED_IN_USER,
                                 'attrkey': 'foo',
                                 'attrvalue': 'baz'})
        with self.app.test_request_context('/token/',
                                           method='GET',
                                           headers={'Authorization': user_token}):
            res = self.app.full_dispatch_request()
            self.assertTrue(res.status_code == 200, res)
            self.assertEqual('baz', user.attributes['foo'], user.attributes)

        delete_event(eid)
        user.delete_attribute('foo')


from privacyidea.lib.smtpserver import add_smtpserver
from . import smtpmock


class EventWrapperTestCase(MyApiTestCase):
    # Test the wrapper/decorator in lib/event.py
    # In other cases we test specific event handlers, but not the calling of the event handler
    # We do this here via an API call.
    serial = "myToken"

    def test_00_setup(self):
        # setup realms
        self.setUp_user_realms()

        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

    @smtpmock.activate
    def test_01_sendmail_post(self):
        r = set_event("send email", "token_init", "UserNotification", "sendmail",
                      conditions={},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com"})
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        with mock.patch("logging.Logger.warning") as mock_log:
            with self.app.test_request_context('/token/init',
                                               data={"genkey": 1,
                                                     "serial": self.serial},
                                               headers={'Authorization': self.at},
                                               method='POST'):
                res = self.app.full_dispatch_request()
                self.assertTrue(res.status_code == 200, res)
                result = res.json.get("result")
                self.assertTrue(result.get("value"), result)
            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['pretzel@example.com']}"
            mock_log.assert_called_once_with(expected)
            msg = smtpmock.get_sent_message()
            self.assertIn('To: pretzel@example.com', msg)
        delete_event(r)

    @smtpmock.activate
    def test_02_sendmail_pre(self):
        r = set_event("send email", "token_init", "UserNotification", "sendmail",
                      conditions={},
                      position="pre",
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "donut@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com"})
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"donut@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        with mock.patch("logging.Logger.warning") as mock_log:
            with self.app.test_request_context('/token/init',
                                               data={"genkey": 1,
                                                     "serial": self.serial},
                                               headers={'Authorization': self.at},
                                               method='POST'):
                res = self.app.full_dispatch_request()
                self.assertTrue(res.status_code == 200, res)
                result = res.json.get("result")
                self.assertTrue(result.get("value"), result)
            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['donut@example.com']}"
            mock_log.assert_called_once_with(expected)

            msg = smtpmock.get_sent_message()
            self.assertIn('To: donut@example.com', msg)
        delete_event(r)


class ContainerHandlerTestCase(MyApiTestCase):
    """
    Full workflow tests for container actions:
    These tests cover some typical use cases for event handlers, but not all possible combinations
    """

    def request_assert_success(self, url, data: dict, auth_token, method='POST'):
        with self.app.test_request_context(url,
                                           method=method,
                                           data=data if method == 'POST' else None,
                                           query_string=data if method == 'GET' else None,
                                           headers={'Authorization': auth_token} if auth_token else None):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res.json)
            self.assertTrue(res.json["result"]["status"])
        return res.json

    def request_assert_error(self, url, data: dict, auth_token, method='POST'):
        with self.app.test_request_context(url,
                                           method=method,
                                           data=data if method == 'POST' else None,
                                           query_string=data if method == 'GET' else None,
                                           headers={'Authorization': auth_token} if auth_token else None):
            res = self.app.full_dispatch_request()
            self.assertFalse(res.json["result"]["status"])
        return res.json

    @classmethod
    def setup_container_with_tokens(cls):
        # create container
        container_serial = init_container({"type": "smartphone"})["container_serial"]
        container = find_container_by_serial(container_serial)

        # create tokens
        hotp = init_token({"type": "hotp", "genkey": True})
        container.add_token(hotp)
        totp = init_token({"type": "totp", "genkey": True})
        container.add_token(totp)

        return container, hotp, totp

    def test_01_disable_tokens_on_unregister(self):
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        # check that tokens are enabled
        self.assertTrue(hotp.is_active())
        self.assertTrue(totp.is_active())

        # Register container
        ContainerEventTestCase.register_smartphone(container)

        # Create event handler definition
        eid = set_event("event", event=["container_unregister"], handlermodule="Container",
                        action=C_ACTION_TYPE.DISABLE_TOKENS, conditions={}, active=True, position="post")

        # Unregister
        self.request_assert_success(f'container/register/{container.serial}/terminate',
                                    {},
                                    self.at, 'POST')

        # Check that tokens are disabled
        self.assertFalse(hotp.is_active())
        self.assertFalse(totp.is_active())

        # Clean up
        delete_event(eid)
        container.delete()
        hotp.delete_token()
        totp.delete_token()

    def test_02_disable_enable_tokens_on_rollover(self):
        # Disable all tokens if a rollover is initiated and enable them after the rollover is successfully completed
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        # check that tokens are enabled
        self.assertTrue(hotp.is_active())
        self.assertTrue(totp.is_active())

        # Register container
        ContainerEventTestCase.register_smartphone(container)

        # Create event handler definition
        eid_disable = set_event("disable", event=["container_register_initialize"], handlermodule="Container",
                                action=C_ACTION_TYPE.DISABLE_TOKENS, conditions=None,
                                ordering=None, options=None, active=True,
                                position="post")
        eid_enable = set_event("enable", event=["container_register_finalize"], handlermodule="Container",
                               action=C_ACTION_TYPE.ENABLE_TOKENS, conditions=None,
                               ordering=None, options=None, active=True,
                               position="post")

        # Init rollover
        set_policy("policy", scope=SCOPE.CONTAINER, action={ACTION.PI_SERVER_URL: "https://pi.net/"}, priority=2)
        result = self.request_assert_success(f'container/register/initialize',
                                             {"container_serial": container.serial, "rollover": True},
                                             self.at, 'POST')
        init_result = result["result"]["value"]

        # Check that tokens are disabled
        self.assertFalse(hotp.is_active())
        self.assertFalse(totp.is_active())

        # Finalize rollover fails
        self.request_assert_error(f'container/register/finalize',
                                  {"container_serial": container.serial},
                                  self.at, 'POST')
        # Check that tokens are disabled
        self.assertFalse(hotp.is_active())
        self.assertFalse(totp.is_active())

        # Finalize rollover success
        scope = f"https://pi.net/container/register/finalize"
        mock_smph = MockSmartphone()
        params = mock_smph.register_finalize(init_result["nonce"], init_result["time_stamp"],
                                             scope, container.serial)

        self.request_assert_success(f'container/register/finalize',
                                    params,
                                    self.at, 'POST')

        # check that tokens are enabled
        self.assertTrue(hotp.is_active())
        self.assertTrue(totp.is_active())

        # Clean up
        delete_policy("policy")
        delete_event(eid_enable)
        delete_event(eid_disable)
        container.delete()
        hotp.delete_token()
        totp.delete_token()

    def test_03_disable_tokens_if_one_is_locked(self):
        # Disable all tokens if one token in the container is locked
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        # check that tokens are enabled
        self.assertTrue(hotp.is_active())
        self.assertTrue(totp.is_active())

        # setup token
        hotp.set_maxfail(1)

        # set event handler definition
        eid = set_event("disable", event=["validate_check"], handlermodule="Container",
                        action=C_ACTION_TYPE.DISABLE_TOKENS, conditions={"token_locked": "True"},
                        ordering=None, options=None, active=True,
                        position="post")

        # login with token
        self.request_assert_success(f'/validate/check',
                                    {"serial": hotp.get_serial(),
                                     "pass": "1234"},
                                    None, 'POST')

        # check that tokens are disabled
        self.assertFalse(hotp.is_active())
        self.assertFalse(totp.is_active())

        # Cleanup
        delete_event(eid)
        container.delete()
        hotp.delete_token()
        totp.delete_token()

    @smtpmock.activate
    def test_04_mail_on_register_init(self):
        """
        Send a mail with the registration QR code to the user.
        """
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()
        # policy for registration
        set_policy("policy", scope=SCOPE.CONTAINER, action={ACTION.PI_SERVER_URL: "https://pi.net/"})

        self.setUp_user_realms()
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        msg = """
                Hello {user},

                the administrator {admin}@{realm} initialized a registration for your container {container_serial}.
                To complete the registration, please scan the attached QR code or click on this link 
                {container_url_value} on your smartphone.
                
                The QR code allows you to synchronize the tokens on your smartphone with the server. 
                Be aware of no one else can see this QR code, otherwise your tokens might be compromised. 
                
                <img src={container_url_img}>

                To check your container you may login to the Web UI:
                {url}
                """
        r = set_event("send email", "container_register_initialize", "UserNotification", "sendmail",
                      conditions={},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com",
                               "subject": "Container Registration.",
                               "body": msg,
                               "attach_qrcode": True,
                               "mimetype": "html"})
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        with mock.patch("logging.Logger.warning") as mock_log:
            self.request_assert_success(f'/container/register/initialize',
                                        {"container_serial": container.serial},
                                        self.at, 'POST')

            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['pretzel@example.com']}"
            mock_log.assert_called_once_with(expected)
            msg = smtpmock.get_sent_message()
            self.assertIn('To: pretzel@example.com', msg)
            self.assertIn("Container Registration", msg)
        delete_event(r)
        delete_policy("policy")

    @smtpmock.activate
    def test_05_mail_on_unregister(self):
        """
        Send a user notification if the container is unregistered
        The smtp server is mocked to return an error. The event is successfully triggered if the error occurs in the
        log.
        """
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        # Register container
        ContainerEventTestCase.register_smartphone(container)

        self.setUp_user_realms()
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        msg = """
                Hello {user},
                
                the administrator {admin}@{realm} performed the action
                {action} on your container {container_serial}.
                
                To check your tokens you may login to the Web UI:
                {url}
                """
        r = set_event("send email", "container_unregister", "UserNotification", "sendmail",
                      conditions={},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com",
                               "subject": "Your container was unregistered.",
                               "body": msg,
                               "mimetype": "html"})
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        with mock.patch("logging.Logger.warning") as mock_log:
            self.request_assert_success(f'/container/register/{container.serial}/terminate',
                                        {},
                                        self.at, 'POST')

            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['pretzel@example.com']}"
            mock_log.assert_called_once_with(expected)
            msg = smtpmock.get_sent_message()
            self.assertIn('To: pretzel@example.com', msg)
            self.assertIn("Your container was unregistered.", msg)
        delete_event(r)

    @smtpmock.activate
    def test_06_mail_on_client_rollover(self):
        """
        Send a user notification if the container shall be rolled over
        """
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        self.setUp_user_realms()
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        msg = """
                Hello {user},

                a rollover was requested for your container {container_serial}.
                The request came from the IP {client_ip}.
                If that was you, you can ignore this mail, otherwise your container and all tokens might be compromised.
                Please contact your admin and deactivate all tokens in the Web UI {url}.
                """
        r = set_event("send email", "container_init_rollover", "UserNotification", "sendmail",
                      conditions={},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com",
                               "subject": "Container Rollover.",
                               "body": msg,
                               "mimetype": "html"},
                      position="post")
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        # Prepare rollover
        # Register container
        mock_smph = ContainerEventTestCase.register_smartphone(container)
        set_policy("policy", scope=SCOPE.CONTAINER,
                   action={ACTION.PI_SERVER_URL: "https://pi.net/", ACTION.CONTAINER_CLIENT_ROLLOVER: True})
        # Create Challenge for rollover
        scope = "https://pi.net/container/rollover"
        challenge_data = container.create_challenge(scope)
        # Mock smartphone
        params = mock_smph.register_finalize(challenge_data["nonce"],
                                             challenge_data["time_stamp"],
                                             scope)

        with mock.patch("logging.Logger.warning") as mock_log:
            self.request_assert_success("/container/rollover",
                                        params,
                                        None, 'POST')

            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['pretzel@example.com']}"
            mock_log.assert_called_once_with(expected)
            msg = smtpmock.get_sent_message()
            self.assertIn('To: pretzel@example.com', msg)
            self.assertIn("Container Rollover", msg)
        delete_event(r)
        delete_policy("policy")

    @smtpmock.activate
    def test_07_mail_on_rollover(self):
        """
        Send a user notification if the container shall be rolled over
        """
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        self.setUp_user_realms()
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        msg = """
                Hello {user},

                the administrator {admin}@{realm} initialized a rollover for your container {container_serial}.
                To complete the rollover, please scan the attached QR code or click on this link {container_url_value}
                on your smartphone.
                
                The QR code allows you to synchronize the tokens on your smartphone with the server. 
                Be aware of no one else can see this QR code, otherwise your tokens might be compromised. 
                
                <img src={container_url_img}>
                
                After the rollover is completed, the tokens and container on the old device are not valid anymore.

                To check your container you may login to the Web UI:
                {url}
                """
        r = set_event("send email", "container_register_initialize", "UserNotification", "sendmail",
                      conditions={"container_info": "registration_state==rollover"},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com",
                               "subject": "Container Rollover.",
                               "body": msg,
                               "mimetype": "html"},
                      position="post")
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        set_policy("policy", scope=SCOPE.CONTAINER,
                   action={ACTION.PI_SERVER_URL: "https://pi.net/", ACTION.CONTAINER_CLIENT_ROLLOVER: True})

        # Register container
        ContainerEventTestCase.register_smartphone(container)

        with mock.patch("logging.Logger.warning") as mock_log:
            self.request_assert_success(f'/container/register/initialize',
                                        {"container_serial": container.serial, "rollover": True},
                                        self.at, 'POST')

            # Check warning in log
            expected = "Failed to send a notification email to user {'email': ['pretzel@example.com']}"
            mock_log.assert_called_once_with(expected)
            msg = smtpmock.get_sent_message()
            self.assertIn('To: pretzel@example.com', msg)
            self.assertIn("Container Rollover", msg)

        delete_event(r)
        delete_policy("policy")

    @smtpmock.activate
    def test_08_mail_on_rollover_not_triggered_for_register(self):
        """
        Send a user notification if the container shall be rolled over.
        The mail shall not be triggered, if a container is initially registered.
        """
        # create container with tokens
        container, hotp, totp = self.setup_container_with_tokens()

        self.setUp_user_realms()
        r = add_smtpserver(identifier="myserver", server="1.2.3.4", tls=False)
        self.assertTrue(r > 0)

        r = set_event("send email", "container_register_initialize", "UserNotification", "sendmail",
                      conditions={"container_info": "registration_state==rollover"},
                      options={"emailconfig": "myserver",
                               "To": "email",
                               "To email": "pretzel@example.com",
                               "reply_to": "email",
                               "reply_to email": "privacyidea@example.com",
                               "subject": "Container Rollover.",
                               "body": ""},
                      position="post")
        self.assertTrue(r > 0)

        smtpmock.setdata(response={"pretzel@example.com": (450, "Mailbox not available")},
                         support_tls=False)

        set_policy("policy", scope=SCOPE.CONTAINER,
                   action={ACTION.PI_SERVER_URL: "https://pi.net/", ACTION.CONTAINER_CLIENT_ROLLOVER: True})

        with mock.patch("logging.Logger.warning") as mock_log:
            self.request_assert_success(f'/container/register/initialize',
                                        {"container_serial": container.serial},
                                        self.at, 'POST')

            # mock log not called
            mock_log.assert_not_called()

        delete_event(r)
        delete_policy("policy")



# 2024-05-08 Cornelius Kölbel <cornelius.koelbel@netknights.it>
# SPDX-FileCopyrightText: (C) 2024 Cornelius Kölbel <cornelius.koelbel@netknights.it>
#
# SPDX-License-Identifier: AGPL-3.0-or-later
#
# Info: https://privacyidea.org
#
# This code is free software: you can redistribute it and/or
# modify it under the terms of the GNU Affero General Public License
# as published by the Free Software Foundation, either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program. If not, see <http://www.gnu.org/licenses/>.

import logging
import re

log = logging.getLogger(__name__)


def validate_email(email):
    """
    Verify if the email is valid format
    :param email: email address
    :return: True if valid, False otherwise
    """
    # regular expression for email validation
    regex = r'^[-\w\.]+@([\w-]+\.)+[\w-]{2,4}$'
    if re.search(regex, email):
        return True
    else:
        return False



"""Add the boolean field check_all_resolvers to policy table.

Revision ID: 58e4f7ebb705
Revises: 3f7e8583ea2
Create Date: 2016-12-13 15:20:54.003819

"""

# revision identifiers, used by Alembic.
revision = '58e4f7ebb705'
down_revision = '3f7e8583ea2'

from alembic import op
import sqlalchemy as sa


def upgrade():
    try:
        op.add_column('policy', sa.Column('check_all_resolvers', sa.Boolean(), nullable=True))
    except Exception as exx:
        print ("Could not add column 'check_all_resolvers'")
        print (exx)


def downgrade():
    op.drop_column('policy', 'check_all_resolvers')



"""alter table challenge

Revision ID: 307a4fbe8a05
Revises: 1edda52b619f
Create Date: 2017-04-19 14:39:20.255958

"""

# revision identifiers, used by Alembic.
revision = '307a4fbe8a05'
down_revision = '1edda52b619f'

from alembic import op


def upgrade():
    try:
        op.create_index(op.f('ix_challenge_serial'), 'challenge', ['serial'],
                        unique=False)
    except Exception as exx:
        print("Could not add index to 'challenge.serial'")
        print (exx)
    try:
        op.drop_index('ix_challenge_transaction_id', table_name='challenge')
        op.create_index(op.f('ix_challenge_transaction_id'), 'challenge',
                        ['transaction_id'], unique=False)
    except Exception as exx:
        print("Could not remove uniqueness from 'challenge.transaction_id'")
        print (exx)


def downgrade():
    op.drop_index(op.f('ix_challenge_transaction_id'), table_name='challenge')
    op.create_index('ix_challenge_transaction_id', 'challenge',
                    ['transaction_id'], unique=1)
    op.drop_index(op.f('ix_challenge_serial'), table_name='challenge')



"""Add column Type to the TokenInfo

Revision ID: 2551ee982544
Revises: 4f32a4e1bf33
Create Date: 2015-02-02 18:05:02.480354

"""

# revision identifiers, used by Alembic.
revision = '2551ee982544'
down_revision = '4f32a4e1bf33'

from alembic import op
import sqlalchemy as sa
from sqlalchemy.exc import OperationalError, ProgrammingError, InternalError


def upgrade():
    try:
        op.add_column('tokeninfo', sa.Column('Type', sa.Unicode(length=100),
                                             nullable=True))
    except (OperationalError, ProgrammingError, InternalError) as exx:
        if "duplicate column name" in str(exx.orig).lower():
            print("Good. Column tokeninfo already exists.")
        else:
            print("Column already exists")
            print(exx)

    except Exception as exx:
        print("Could not add the column 'Type' to table tokeninfo")
        print (exx)
    ### end Alembic commands ###


def downgrade():
    op.drop_column('tokeninfo', 'Type')
    ### end Alembic commands ###



# (c) NetKnights GmbH 2024,  https://netknights.it
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# SPDX-FileCopyrightText: 2024 Nils Behlen <nils.behlen@netknights.it>
# SPDX-License-Identifier: AGPL-3.0-or-later
#
from unittest.mock import patch

from webauthn.helpers.structs import AttestationConveyancePreference

import privacyidea.lib.token
from privacyidea.config import TestingConfig
from privacyidea.lib.fido2.policy_action import FIDO2PolicyAction, PasskeyAction
from privacyidea.lib.framework import get_app_config_value
from privacyidea.lib.policy import set_policy, SCOPE, delete_policy
from privacyidea.lib.token import remove_token, init_token, get_tokens
from privacyidea.lib.tokens.webauthn import CoseAlgorithm
from privacyidea.lib.user import User
from tests.base import MyApiTestCase, OverrideConfigTestCase
from tests.passkey_base import PasskeyTestBase


class PasskeyAPITestBase(MyApiTestCase, PasskeyTestBase):

    def setUp(self):
        PasskeyTestBase.setUp(self)
        self.setUp_user_realms()
        self.user = User(login="hans", realm=self.realm1,
                         resolver=self.resolvername1)
        PasskeyTestBase.__init__(self)

        set_policy("passkey_rp_id", scope=SCOPE.ENROLL, action=f"{FIDO2PolicyAction.RELYING_PARTY_ID}={self.rp_id}")
        set_policy("passkey_rp_name", scope=SCOPE.ENROLL,
                   action=f"{FIDO2PolicyAction.RELYING_PARTY_NAME}={self.rp_id}")
        self.pk_headers = {'ORIGIN': self.expected_origin, 'Authorization': self.at}
        # Delete all token
        tokens = get_tokens(user=self.user)
        for t in tokens:
            remove_token(t.get_serial())

    def tearDown(self):
        delete_policy("passkey_rp_id")
        delete_policy("passkey_rp_name")

    def _token_init_step_one(self):
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.registration_challenge

            with self.app.test_request_context('/token/init',
                                               method='POST',
                                               data={"type": "passkey", "user": self.user.login,
                                                     "realm": self.user.realm},
                                               headers=self.pk_headers):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                self.assertIn("detail", res.json)
                detail = res.json["detail"]
                self.assertIn("passkey_registration", detail)
                self.validate_default_passkey_registration(detail["passkey_registration"])
                self.assertIn("rollout_state", detail)
                self.assertEqual("clientwait", detail["rollout_state"])
                passkey_registration = detail["passkey_registration"]
                # PubKeyCredParams: Via the API, all three key algorithms (from webauthn) are valid by default
                self.assertEqual(len(passkey_registration["pubKeyCredParams"]), 3)
                for param in passkey_registration["pubKeyCredParams"]:
                    self.assertIn(param["type"], ["public-key"])
                    self.assertIn(param["alg"], [-7, -37, -257])
                # ExcludeCredentials should be empty because no other passkey token is registered for the user
                self.assertEqual(0, len(passkey_registration["excludeCredentials"]),
                                 "excludeCredentials should be empty")
                return res.json

    def _token_init_step_two(self, transaction_id, serial):
        data = {
            "attestationObject": self.registration_attestation,
            "clientDataJSON": self.registration_client_data,
            "credential_id": self.credential_id,
            "rawId": self.credential_id,
            "authenticatorAttachment": self.authenticator_attachment,
            FIDO2PolicyAction.RELYING_PARTY_ID: self.rp_id,
            "transaction_id": transaction_id,
            "type": "passkey",
            "user": self.user.login,
            "realm": self.user.realm,
            "serial": serial
        }
        with self.app.test_request_context('/token/init', method='POST', data=data, headers=self.pk_headers):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self._assert_result_value_true(res.json)

    def _enroll_static_passkey(self) -> str:
        """
        Returns the serial of the enrolled passkey token
        """
        data = self._token_init_step_one()
        detail = data["detail"]
        serial = detail["serial"]
        transaction_id = detail["transaction_id"]
        self._token_init_step_two(transaction_id, serial)
        return serial

    def _trigger_passkey_challenge(self, mock_nonce: str) -> dict:
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = mock_nonce
            with self.app.test_request_context('/validate/initialize', method='POST', data={"type": "passkey"}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                self.assertIn("detail", res.json)
                detail = res.json["detail"]
                self.assertIn("transaction_id", detail)
                self.assertIn("passkey", detail)
                passkey = detail["passkey"]
                self.assertIn("challenge", passkey)
                self.assertEqual(mock_nonce, passkey["challenge"])
                self.assertIn("message", passkey)
                self.assertIn("transaction_id", passkey)
                self.assertIn("rpId", passkey)
                self.assertEqual(self.rp_id, passkey["rpId"])
            return passkey

    def _assert_result_value_true(self, response_json):
        self.assertIn("result", response_json)
        self.assertIn("status", response_json["result"])
        self.assertTrue(response_json["result"]["status"])
        self.assertIn("value", response_json["result"])
        self.assertTrue(response_json["result"]["value"])

    def _verify_auth_fail_with_error(self, res, error_code: int):
        self.assertEqual(401, res.status_code)
        self.assertIn("result", res.json)
        self.assertIn("status", res.json["result"])
        self.assertFalse(res.json["result"]["status"])
        self.assertIn("error", res.json["result"])
        error = res.json["result"]["error"]
        self.assertIn("message", error)
        self.assertTrue(error["message"])
        self.assertIn("code", error)
        self.assertEqual(error_code, error["code"])


class PasskeyAPITest(PasskeyAPITestBase):
    """
    Passkey uses challenges that are not bound to a user.
    A successful authentication with a passkey should return the username.
    Passkeys can be used with cross-device sign-in, similar to how push token work
    """

    def test01_token_init_with_policies(self):
        # Test if setting the policies alters the registration data correctly
        # Create a passkey token so excludeCredentials is not empty
        serial = self._enroll_static_passkey()

        set_policy("key_algorithm", scope=SCOPE.ENROLL,
                   action=f"{FIDO2PolicyAction.PUBLIC_KEY_CREDENTIAL_ALGORITHMS}=ecdsa")
        set_policy("attestation", scope=SCOPE.ENROLL, action=f"{PasskeyAction.AttestationConveyancePreference}="
                                                             f"{AttestationConveyancePreference.ENTERPRISE.value}")
        set_policy("user_verification", scope=SCOPE.ENROLL,
                   action=f"{FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT}=required")

        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.registration_challenge

            with self.app.test_request_context('/token/init',
                                               method='POST',
                                               data={"type": "passkey", "user": self.user.login,
                                                     "realm": self.user.realm},
                                               headers=self.pk_headers):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                data = res.json
                self.assertIn("detail", data)
                self.assertIn("serial", data["detail"])
                serial_2 = data["detail"]["serial"]
                self.assertIn("passkey_registration", data["detail"])
                passkey_registration = data["detail"]["passkey_registration"]
                # PubKeyCredParams: Only ecdsa should be allowed
                self.assertEqual(len(passkey_registration["pubKeyCredParams"]), 1)
                self.assertEqual(passkey_registration["pubKeyCredParams"][0]["alg"], CoseAlgorithm.ES256)
                # Attestation should be enterprise
                self.assertEqual(passkey_registration["attestation"], AttestationConveyancePreference.ENTERPRISE)
                # ExcludeCredentials should contain the credential id of the registered token
                self.assertEqual(len(passkey_registration["excludeCredentials"]), 1)
                self.assertEqual(passkey_registration["excludeCredentials"][0]["id"], self.credential_id)

        delete_policy("key_algorithm")
        delete_policy("attestation")
        delete_policy("user_verification")
        remove_token(serial)
        remove_token(serial_2)

    def test_02_authenticate_no_uv(self):
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_no_uv)
        self.assertIn("user_verification", passkey_challenge)
        # By default, user_verification is preferred
        self.assertEqual("preferred", passkey_challenge["user_verification"])

        transaction_id = passkey_challenge["transaction_id"]
        # Answer the challenge
        data = self.authentication_response_no_uv
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/validate/check', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self._assert_result_value_true(res.json)
            self.assertNotIn("auth_items", res.json)
        remove_token(serial)

    def test_03_authenticate_wrong_uv(self):
        """
        Wrong UV meaning user verification is required but the authenticator data does not contain the UV flag
        """
        serial = self._enroll_static_passkey()
        set_policy("user_verification", scope=SCOPE.AUTH,
                   action=f"{FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT}=required")
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_no_uv)
        self.assertIn("user_verification", passkey_challenge)
        self.assertEqual("required", passkey_challenge["user_verification"])
        transaction_id = passkey_challenge["transaction_id"]

        data = self.authentication_response_no_uv
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/validate/check', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self.assertIn("result", res.json)
            self.assertIn("status", res.json["result"])
            self.assertTrue(res.json["result"]["status"])
            # Value is false and authentication is REJECT
            self.assertIn("value", res.json["result"])
            self.assertFalse(res.json["result"]["value"])
            self.assertIn("authentication", res.json["result"])
            self.assertEqual("REJECT", res.json["result"]["authentication"])

        remove_token(serial)
        delete_policy("user_verification")

    def test_04_authenticate_with_uv(self):
        serial = self._enroll_static_passkey()
        set_policy("user_verification", scope=SCOPE.AUTH,
                   action=f"{FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT}=required")
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_uv)
        self.assertIn("user_verification", passkey_challenge)
        self.assertEqual("required", passkey_challenge["user_verification"])
        transaction_id = passkey_challenge["transaction_id"]

        data = self.authentication_response_uv
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/validate/check', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self._assert_result_value_true(res.json)

        remove_token(serial)
        delete_policy("user_verification")

    def test_05_trigger_with_pin(self):
        """
        By default, passkeys are not triggered using the PIN, because the flow of authentication is very different
        from our other token types. However, it is possible to enable this behavior with the
        policy passkey_trigger_with_pin.
        """
        # Without the policy, the passkey should not be triggered with the PIN
        serial = self._enroll_static_passkey()
        with self.app.test_request_context('/validate/check', method='POST',
                                           data={"user": self.user.login, "pass": ""}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self.assertEqual("REJECT", res.json["result"]["authentication"])
            self.assertFalse(res.json["result"]["value"])


        # Now set the policy to trigger the passkey with the PIN
        set_policy("passkey_trigger_with_pin", scope=SCOPE.AUTH, action=f"{PasskeyAction.EnableTriggerByPIN}=true")
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.authentication_challenge_no_uv
            with self.app.test_request_context('/validate/check', method='POST',
                                               data={"user": self.user.login, "pass": ""}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                self.assertIn("detail", res.json)
                detail = res.json["detail"]
                self.assertIn("multi_challenge", detail)

                multi_challenge = detail["multi_challenge"]
                self.assertEqual(len(multi_challenge), 1)

                challenge = multi_challenge[0]
                self.assertIn("transaction_id", challenge)
                transaction_id = challenge["transaction_id"]
                self.assertTrue(transaction_id)

                self.assertIn("challenge", challenge)
                self.assertEqual(self.authentication_challenge_no_uv, challenge["challenge"])

                self.assertIn("serial", challenge)
                self.assertEqual(serial, challenge["serial"])

                self.assertIn("type", challenge)
                self.assertEqual("passkey", challenge["type"])

                self.assertIn("userVerification", challenge)
                self.assertTrue(challenge["userVerification"])

                self.assertIn("rpId", challenge)
                self.assertEqual(self.rp_id, challenge["rpId"])

                self.assertIn("message", challenge)
                self.assertTrue(challenge["message"])

                self.assertIn("client_mode", challenge)
                self.assertEqual("webauthn", challenge["client_mode"])

        # Answer the challenge
        data = self.authentication_response_no_uv
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/validate/check', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self._assert_result_value_true(res.json)
            self.assertIn("result", res.json)
            self.assertIn("authentication", res.json["result"])
            self.assertEqual("ACCEPT", res.json["result"]["authentication"])
            # Should return the username
            self.assertIn("detail", res.json)
            detail = res.json["detail"]
            self.assertIn("username", detail)
            self.assertEqual(self.user.login, detail["username"])
            self.assertNotIn("auth_items", res.json)
        remove_token(serial)
        delete_policy("passkey_trigger_with_pin")

    def test_06_validate_check_wrong_serial(self):
        """
        Challenges triggered via /validate/check should be bound to a specific serial.
        Trying to answer the challenge with a token with a different serial should fail.
        """
        set_policy("passkey_trigger_with_pin", scope=SCOPE.AUTH, action=f"{PasskeyAction.EnableTriggerByPIN}=true")
        serial = self._enroll_static_passkey()
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.authentication_challenge_no_uv
            with self.app.test_request_context('/validate/check', method='POST',
                                               data={"user": self.user.login, "pass": ""}):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                detail = res.json["detail"]
                transaction_id = detail["multi_challenge"][0]["transaction_id"]
        # Change the token serial
        token = privacyidea.lib.token.get_tokens(serial=serial)[0]
        token.token.serial = "123456"
        token.token.save()
        # Try to answer the challenge, will fail
        data = self.authentication_response_no_uv
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/validate/check', method='POST', data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code)
            result = res.json["result"]
            self.assertIn("error", result)
            error = result["error"]
            self.assertIn("message", error)
            self.assertIn("code", error)
            self.assertEqual(403, error["code"])
            self.assertFalse(result["status"])
        remove_token(token.token.serial)
        delete_policy("passkey_trigger_with_pin")

    def test_07_trigger_challenge(self):
        """
        Just test if the challenge is returned by /validate/triggerchallenge. The response would be sent to
        /validate/check and that is already tested. Requires the passkey_trigger_with_pin policy to be set.
        """
        set_policy("passkey_trigger_with_pin", scope=SCOPE.AUTH, action=f"{PasskeyAction.EnableTriggerByPIN}=true")
        serial = self._enroll_static_passkey()
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.authentication_challenge_no_uv
            with self.app.test_request_context('/validate/triggerchallenge', method='POST',
                                               data={"user": self.user.login}, headers=self.pk_headers):
                res = self.app.full_dispatch_request()
                self.assertEqual(200, res.status_code)
                self.assertIn("detail", res.json)
                detail = res.json["detail"]
                self.assertIn("multi_challenge", detail)

                multi_challenge = detail["multi_challenge"]
                self.assertEqual(len(multi_challenge), 1)

                challenge = multi_challenge[0]
                self.assertIn("transaction_id", challenge)
                transaction_id = challenge["transaction_id"]
                self.assertTrue(transaction_id)

                self.assertIn("challenge", challenge)
                self.assertEqual(self.authentication_challenge_no_uv, challenge["challenge"])

                self.assertIn("serial", challenge)
                self.assertEqual(serial, challenge["serial"])

                self.assertIn("type", challenge)
                self.assertEqual("passkey", challenge["type"])

                self.assertIn("userVerification", challenge)
                self.assertTrue(challenge["userVerification"])

                self.assertIn("rpId", challenge)
                self.assertEqual(self.rp_id, challenge["rpId"])

                self.assertIn("message", challenge)
                self.assertTrue(challenge["message"])

                self.assertIn("client_mode", challenge)
                self.assertEqual("webauthn", challenge["client_mode"])
        remove_token(serial)

    def test_08_offline(self):
        serial = self._enroll_static_passkey()
        data = {"serial": serial, "machineid": 0, "application": "offline", "resolver": ""}
        with self.app.test_request_context('/machine/token', method='POST', data=data, headers=self.pk_headers):
            res = self.app.full_dispatch_request()
            self.assertIn("result", res.json)
            self.assertIn("status", res.json["result"])
            self.assertTrue(res.json["result"]["status"])
            self.assertIn("value", res.json["result"])
            self.assertEqual(1, res.json["result"]["value"])

        # A successful authentication should return the offline data now
        challenge = self._trigger_passkey_challenge(self.authentication_challenge_no_uv)
        transaction_id = challenge["transaction_id"]
        data = self.authentication_response_no_uv
        data["transaction_id"] = transaction_id
        # A user agent that identifies the machine is required.
        user_agent = "privacyidea-cp/1.1.1 Windows/Laptop-1231312"
        headers = {"Origin": self.expected_origin, "User-Agent": user_agent}
        # IP is needed to get the offline data
        with self.app.test_request_context('/validate/check',
                                           method='POST',
                                           environ_base={'REMOTE_ADDR': '10.0.0.17'},
                                           data=data,
                                           headers=headers):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self._assert_result_value_true(res.json)
            self.assertIn("detail", res.json)
            self.assertIn("auth_items", res.json)
            auth_items = res.json["auth_items"]
            self.assertIn("offline", auth_items)
            # Offline items for 1 token should be returned
            offline = auth_items["offline"]
            self.assertEqual(1, len(offline))
            offline = offline[0]
            # At least for this test user=username
            self.assertIn("user", offline)
            self.assertEqual(self.user.login, offline["user"])
            self.assertIn("refilltoken", offline)
            self.assertTrue(offline["refilltoken"])
            refill_token = offline["refilltoken"]
            self.assertIn("username", offline)
            self.assertEqual(self.user.login, offline["username"])
            self.assertIn("response", offline)
            response = offline["response"]
            self.assertIn("rpId", response)
            self.assertEqual(self.rp_id, response["rpId"])
            self.assertIn("pubKey", response)
            self.assertIn("credentialId", response)
            # Verify that the returned values are correct
            token = privacyidea.lib.token.get_tokens(serial=serial)[0]
            public_key = token.get_tokeninfo("public_key")
            self.assertEqual(public_key, response["pubKey"])
            credential_id = token.token.get_otpkey().getKey().decode("utf-8")
            self.assertEqual(credential_id, response["credentialId"])

        # Refill without machine name will fail with parameter error 905
        data = {"serial": serial, "refilltoken": refill_token, "pass": ""}
        with self.app.test_request_context('/validate/offlinerefill', method='POST', data=data):
            res = self.app.full_dispatch_request()
            self.assertIn("result", res.json)
            result = res.json["result"]
            self.assertIn("status", result)
            self.assertFalse(result["status"])
            self.assertIn("error", result)
            self.assertIn("message", result["error"])
            self.assertTrue(result["error"]["message"])
            self.assertIn("code", result["error"])
            self.assertEqual(905, result["error"]["code"])

        # Refill with machine name works
        data = {"serial": serial, "refilltoken": refill_token, "pass": ""}
        with self.app.test_request_context('/validate/offlinerefill', method='POST', data=data,
                                           headers=headers):
            res = self.app.full_dispatch_request()
            self._assert_result_value_true(res.json)
            # For FIDO2 offline, the refill just checks if the token is still marked as offline and returns
            # a new refill token
            self.assertIn("auth_items", res.json)
            auth_items = res.json["auth_items"]
            self.assertIn("offline", auth_items)
            offline = auth_items["offline"]
            self.assertEqual(1, len(offline))
            offline = offline[0]
            self.assertIn("refilltoken", offline)
            self.assertTrue(offline["refilltoken"])
            self.assertNotEqual(refill_token, offline["refilltoken"])
            refill_token = offline["refilltoken"]
            self.assertIn("serial", offline)
            self.assertEqual(serial, offline["serial"])
            self.assertIn("response", offline)
            self.assertFalse(offline["response"])

        # Disable offline for the token
        with self.app.test_request_context(f'/machine/token/{serial}/offline/1',
                                           method='DELETE',
                                           headers=self.pk_headers):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code)
            self.assertIn("result", res.json)
            self.assertIn("status", res.json["result"])
            self.assertTrue(res.json["result"]["status"])
            self.assertIn("value", res.json["result"])
            self.assertEqual(1, res.json["result"]["value"])

        # Try to refill again, should indicate that the token is no longer valid for offline use.
        data.update({"refilltoken": refill_token})
        with self.app.test_request_context('/validate/offlinerefill', method='POST', data=data,
                                           headers=self.pk_headers):
            res = self.app.full_dispatch_request()
            self.assertIn("result", res.json)
            result = res.json["result"]
            self.assertIn("status", result)
            self.assertFalse(result["status"])
            self.assertIn("error", result)
            error = result["error"]
            self.assertIn("message", error)
            self.assertTrue(error["message"])
            self.assertIn("code", error)
            self.assertEqual(905, error["code"])
        remove_token(serial)

    def test_09_enroll_via_multichallenge(self):
        spass_token = init_token({"type": "spass", "pin": "1"}, self.user)
        action = "enroll_via_multichallenge=PASSKEY, enroll_via_multichallenge_text=enrollVia multichallenge test text"
        set_policy("enroll_passkey", scope=SCOPE.AUTH, action=action)

        # Using the spass token should result in a challenge to enroll a passkey
        with patch('privacyidea.lib.fido2.challenge.get_fido2_nonce') as get_nonce:
            get_nonce.return_value = self.registration_challenge
            with self.app.test_request_context('/validate/check', method='POST',
                                               data={"user": self.user.login, "pass": "1"}):
                res = self.app.full_dispatch_request()
                # Authentication is not successful, instead, it is a challenge
                self.assertIn("result", res.json)
                result = res.json["result"]
                self.assertIn("status", result)
                self.assertTrue(result["status"])
                self.assertIn("value", result)
                self.assertFalse(result["value"], res.json)
                self.assertIn("authentication", result)
                self.assertEqual("CHALLENGE", result["authentication"])
                # Detail
                self.assertIn("detail", res.json)
                detail = res.json["detail"]
                self.assertIn("multi_challenge", detail)
                self.assertIn("client_mode", detail)
                self.assertEqual("webauthn", detail["client_mode"])
                self.assertIn("message", detail)
                self.assertTrue(detail["message"])
                self.assertIn("serial", detail)
                passkey_serial = detail["serial"]
                self.assertTrue(passkey_serial)
                self.assertIn("type", detail)
                self.assertEqual("passkey", detail["type"])
                # Multi challenge
                multi_challenge = detail["multi_challenge"]
                self.assertEqual(1, len(multi_challenge))
                challenge = multi_challenge[0]
                self.assertIn("transaction_id", challenge)
                transaction_id = challenge["transaction_id"]
                self.assertTrue(transaction_id)
                self.assertIn("serial", challenge)
                self.assertTrue(challenge["serial"])
                # Passkey registration
                self.assertIn("passkey_registration", challenge)
                passkey_registration = challenge["passkey_registration"]
                self.assertIn("rp", passkey_registration)
                rp = passkey_registration["rp"]
                self.assertIn("name", rp)
                self.assertEqual(self.rp_id, rp["name"])
                self.assertIn("id", rp)
                self.assertEqual(self.rp_id, rp["id"])
                self.assertIn("user", passkey_registration)
                user = passkey_registration["user"]
                self.assertIn("id", user)
                self.assertIn("name", user)
                self.assertEqual(self.user.login, user["name"])
                self.assertIn("displayName", user)
                self.assertEqual(self.user.login, user["displayName"])
                self.assertIn("challenge", passkey_registration)
                self.assertEqual(self.registration_challenge, passkey_registration["challenge"])
                self.assertIn("pubKeyCredParams", passkey_registration)
                self.assertEqual(3, len(passkey_registration["pubKeyCredParams"]))
                for param in passkey_registration["pubKeyCredParams"]:
                    self.assertIn("type", param)
                    self.assertEqual("public-key", param["type"])
                    self.assertIn("alg", param)
                    self.assertIn(param["alg"], [-7, -37, -257])
                self.assertIn("timeout", passkey_registration)
                self.assertIn("excludeCredentials", passkey_registration)
                self.assertEqual(0, len(passkey_registration["excludeCredentials"]))
                self.assertIn("authenticatorSelection", passkey_registration)
                authenticator_selection = passkey_registration["authenticatorSelection"]
                self.assertIn("residentKey", authenticator_selection)
                self.assertEqual("required", authenticator_selection["residentKey"])
                self.assertIn("requireResidentKey", authenticator_selection)
                self.assertTrue(authenticator_selection["requireResidentKey"])
                self.assertIn("userVerification", authenticator_selection)
                self.assertEqual("preferred", authenticator_selection["userVerification"])
                self.assertIn("attestation", passkey_registration)
                self.assertEqual("none", passkey_registration["attestation"])

        # Answer the challenge
        data = {
            "attestationObject": self.registration_attestation,
            "clientDataJSON": self.registration_client_data,
            "credential_id": self.credential_id,
            "rawId": self.credential_id,
            "authenticatorAttachment": self.authenticator_attachment,
            FIDO2PolicyAction.RELYING_PARTY_ID: self.rp_id,
            "transaction_id": transaction_id,
            "type": "passkey",
            "user": self.user.login,
            "realm": self.user.realm,
            "serial": passkey_serial
        }
        with self.app.test_request_context('/validate/check', method='POST', data=data, headers=self.pk_headers):
            res = self.app.full_dispatch_request()
            self._assert_result_value_true(res.json)
            self.assertIn("detail", res.json)
            detail = res.json["detail"]
            self.assertIn("message", detail)
            self.assertTrue(detail["message"])
            self.assertIn("username", detail)
            self.assertEqual(self.user.login, detail["username"])
            self.assertIn("serial", detail)
            self.assertEqual(passkey_serial, detail["serial"])
            self.assertEqual("ACCEPT", res.json["result"]["authentication"])

        remove_token(spass_token.get_serial())
        remove_token(passkey_serial)
        delete_policy("enroll_passkey")

    def test_10_auth_success(self):
        """
        To use a passkey with /auth, the challenge is initiated via /validate/initialize and
        then answered via /validate/check.
        UserVerification is always required for /auth.
        """
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_uv)
        data = self.authentication_response_uv
        transaction_id = passkey_challenge["transaction_id"]
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/auth', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(200, res.status_code, res.json)
            self.assertIn("result", res.json)
            self.assertIn("status", res.json["result"])
            self.assertTrue(res.json["result"]["status"])
            self.assertIn("value", res.json["result"])
            # Value contains the user data (rights etc.)
            value = res.json["result"]["value"]
            self.assertIn("realm", value)
            self.assertEqual(self.user.realm, value["realm"])
            self.assertIn("log_level", value)
            self.assertIn("menus", value)
            self.assertIn("rights", value)
            self.assertTrue(len(value["rights"]) > 0)
            self.assertIn("role", value)
            self.assertEqual("user", value["role"])
            self.assertIn("token", value)
            self.assertTrue(value["token"])
        remove_token(serial)

    def test_11_auth_fail_uv(self):
        """
        Test an authentication with a wrong response and without user verification.
        """
        set_policy("user_verification", scope=SCOPE.AUTH,
                   action=f"{FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT}=required")
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_no_uv)
        data = self.authentication_response_no_uv
        transaction_id = passkey_challenge["transaction_id"]
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/auth', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self._verify_auth_fail_with_error(res, 4031)
        delete_policy("user_verification")
        remove_token(serial)

    def test_12_auth_fail_signature(self):
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_uv)
        data = self.authentication_response_uv
        transaction_id = passkey_challenge["transaction_id"]
        data["transaction_id"] = transaction_id
        # Change the signature
        data["signature"] = "wrong_signature"
        with self.app.test_request_context('/auth', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self._verify_auth_fail_with_error(res, 4031)
        remove_token(serial)

    def test_13_uv_in_challenge_data(self):
        """
        The user_verification requirement is set in the challenge data. When a challenge is triggered, the value of the
        policy at that moment is stored in the challenge data and used for the authentication.
        If the policy value was changed in the meantime, the authentication will still work because the value from the
        challenge data is used and not read from the policy.
        """
        # Trigger a challenge with no uv requirement, then change it to required and answer the challenge
        # That should still work
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_uv)
        data = self.authentication_response_uv
        transaction_id = passkey_challenge["transaction_id"]
        data["transaction_id"] = transaction_id
        set_policy("user_verification", scope=SCOPE.AUTH,
                   action=f"{FIDO2PolicyAction.USER_VERIFICATION_REQUIREMENT}=required")
        with self.app.test_request_context('/auth', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self._assert_result_value_true(res.json)
        remove_token(serial)
        delete_policy("user_verification")


class PasskeyAuthAPITest(PasskeyAPITestBase, OverrideConfigTestCase):
    """
    Test if the feature switch for passkey usage with /auth works.
    Successful use has been tested before in PasskeyAPITest.
    """

    class Config(TestingConfig):
        WEBUI_PASSKEY_LOGIN_ENABLED = False

    def test_01_webui_passkey_disabled(self):
        self.assertFalse(get_app_config_value("WEBUI_PASSKEY_LOGIN_ENABLED", True))
        serial = self._enroll_static_passkey()
        passkey_challenge = self._trigger_passkey_challenge(self.authentication_challenge_uv)
        data = self.authentication_response_uv
        transaction_id = passkey_challenge["transaction_id"]
        data["transaction_id"] = transaction_id
        with self.app.test_request_context('/auth', method='POST',
                                           data=data,
                                           headers={"Origin": self.expected_origin}):
            res = self.app.full_dispatch_request()
            self.assertEqual(401, res.status_code, res.json)
            self._verify_auth_fail_with_error(res, 4307)
        remove_token(serial)



"""
This test file tests the lib.tokens.4eyestoken
This depends on lib.tokenclass
"""

from .base import MyTestCase
from privacyidea.lib.tokens.foureyestoken import FourEyesTokenClass
from privacyidea.lib.token import init_token, check_serial_pass, remove_token
from privacyidea.lib.user import User


class FourEyesTokenTestCase(MyTestCase):

    def test_00_create_realms(self):
        self.setUp_user_realms()

    def test_01_convert_realm(self):
        r = FourEyesTokenClass.convert_realms("realm1:2,realm2:1")
        self.assertEqual(r, {"realm1": 2,
                             "realm2": 1})

        r = FourEyesTokenClass.convert_realms("realm1,realm2:1")
        self.assertEqual(r, {"realm2": 1})

    def test_02_create_token(self):
        r = init_token({"4eyes": "realm1:1,realm2:2",
                        "type": "4eyes",
                        "separator": "|"})
        self.assertEqual(r.type, "4eyes")
        self.assertEqual(r.get_tokeninfo("4eyes"), "realm1:1,realm2:2")
        self.assertEqual(r.get_tokeninfo("separator"), "|")

        realms = r._get_realms()
        self.assertEqual(realms, {"realm1": 1, "realm2": 2})

        self.assertEqual(r._get_separator(), "|")

    def test_03_authenticate(self):
        self.setUp_user_realms()

        init_token({"type": "pw",
                    "otpkey": "password1",
                    "pin": "pin1",
                    "serial": "pwserial1"},
                   user=User("cornelius", self.realm1))

        init_token({"type": "pw",
                    "otpkey": "password2",
                    "pin": "pin2",
                    "serial": "pwserial2"},
                   user=User("cornelius", self.realm1))

        init_token({"type": "pw",
                    "otpkey": "password3",
                    "pin": "pin3",
                    "serial": "pwserial3"},
                   user=User("cornelius", self.realm1))

        tok = init_token({"serial": "eye1",
                          "type": "4eyes",
                          "4eyes": "{0!s}:2".format(self.realm1),
                          "separator": " "})

        r = check_serial_pass("eye1", "pin1password1 pin2password2")
        self.assertEqual(r[0], True)

        # This triggers the challenge for the next token
        r = check_serial_pass("eye1", "pin1password1")
        self.assertEqual(r[0], False)
        self.assertTrue("transaction_id" in r[1])
        self.assertEqual(r[1].get("message"), 'Please authenticate with another token from either realm: realm1.')

        # check false separator
        r = check_serial_pass("eye1", "pin1password1:pin2password2")
        self.assertFalse(r[0])
        self.assertEqual(r[1].get("foureyes"), "Only found 0 tokens in realm {0!s}".format(self.realm1))

        # check authentication also works if the 4eyes-token is in the same realm
        tok.add_user(User('cornelius', self.realm1))
        r = check_serial_pass("eye1", "pin3password3 pin2password2")
        self.assertTrue(r[0])
        self.assertEqual(r[1].get('message'), 'matching 1 tokens')

        # cleanup
        remove_token(serial='pwserial1')
        remove_token(serial='pwserial2')
        remove_token(serial='pwserial3')
        remove_token(serial='eye1')



#  2018-08-08   Friedrich Weber <friedrich.weber@netknights.it>
#               Add a lifecycle module that allows to
#               register functions to be called after the request
#               has been handled.
#
# This code is free software; you can redistribute it and/or
# modify it under the terms of the GNU AFFERO GENERAL PUBLIC LICENSE
# License as published by the Free Software Foundation; either
# version 3 of the License, or any later version.
#
# This code is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU AFFERO GENERAL PUBLIC LICENSE for more details.
#
# You should have received a copy of the GNU Affero General Public
# License along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
import logging
from privacyidea.lib.framework import get_request_local_store

log = logging.getLogger(__name__)


def register_finalizer(func):
    """
    Register ``func`` to be called after the request has ended (this includes
    cases in which an error has been thrown)
    :param func: a function that takes no arguments
    """
    # from http://flask.pocoo.org/snippets/53/
    store = get_request_local_store()
    if 'call_on_teardown' not in store:
        store['call_on_teardown'] = []
    store['call_on_teardown'].append(func)


def call_finalizers():
    """
    Call all finalizers that have been registered with the current request.
    Exceptions will be caught and written to the log.
    """
    store = get_request_local_store()
    if 'call_on_teardown' in store:
        for func in store['call_on_teardown']:
            try:
                func()
            except Exception as exx:
                log.warning("Caught exception in finalizer: {!r}".format(exx))
                log.debug("Exception in finalizer:", exc_info=True)
        store['call_on_teardown'] = []
